unit UTest1;

interface

uses
  TestFramework, JvInterpreter;

type
  TTest1 = class(TTestCase)
    procedure Test_MethodOrder;
    procedure Test_MethodOrderOk;
  end;

implementation

uses
  System.SysUtils, System.Classes, Vcl.Forms, Vcl.Controls, Winapi.Windows;

{ TTest1 }

type
  TJvAdapterAccess = class(TJvInterpreterAdapter);
  TJvMethodAccess  = class(TJvInterpreterMethod);
  TMyForm          = class(TForm);
  TMyGrid          = class(TWinControl);
  TMyChart         = class(TWinControl);

function DumpList(list: TJvInterpreterIdentifierList): string;
var
  i: Integer;
begin
  Result := '';
  for i := 0 to list.Count - 1 do
    Result := Result + TJvMethodAccess(list.Items[i]).FClassType.ClassName + sLineBreak;
end;

procedure TTest1.Test_MethodOrder;
var
  itpt: TJvInterpreterProgram;
  s1  : string;
begin
  itpt := TJvInterpreterProgram.Create(nil);
  try
    TJvAdapterAccess(itpt.SharedAdapter).Clear;
    itpt.SharedAdapter.AddGet(TMyForm, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TForm, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TMyChart, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TMyGrid, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TScrollingWinControl, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TWinControl, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TControl, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TComponent, 'Method', nil, 0, [], varEmpty);

    TJvInterpreterMethodList(TJvAdapterAccess(itpt.SharedAdapter.GetList)).Sort;
    s1 := DumpList(itpt.SharedAdapter.GetList);
    Check(Pos('TMyForm', s1) < Pos('TScrollingWinControl', s1));

  finally
    itpt.Free;
  end;
end;

function SortMethodIdentifier(Item1, Item2: Pointer): Integer;
begin
  Result := AnsiStrIComp(PChar(TJvInterpreterIdentifier(Item1).Identifier),
      PChar(TJvInterpreterIdentifier(Item2).Identifier));

  if (Result = 0) and (Item1 <> Item2) then
  begin
    if TJvMethodAccess(Item1).FClassType.InheritsFrom(TJvMethodAccess(Item2).FClassType) then
      Result := -1
    else
      Result := 1;
  end;
end;

// copy from System.Classes
procedure QuickSort(SortList: TPointerList; L, R: Integer;
    SCompare: TListSortCompareFunc);
var
  I, J: Integer;
  P, T: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := SortList[(L + R) shr 1];
    repeat
      while SCompare(SortList[I], P) < 0 do
        Inc(I);
      while SCompare(SortList[J], P) > 0 do
        Dec(J);
      if I <= J then
      begin
        if I <> J then
        begin
          T := SortList[I];
          SortList[I] := SortList[J];
          SortList[J] := T;
        end;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then
      QuickSort(SortList, L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure MySort(list: TJvInterpreterIdentifierList);
begin
  // copy from TList.Sort
  if list.Count > 1 then
    QuickSort(list.list, 0, list.Count - 1,
      function(Item1, Item2: Pointer): Integer
      begin
        Result := SortMethodIdentifier(Item1, Item2);
      end);
end;

procedure TTest1.Test_MethodOrderOk;
var
  itpt  : TJvInterpreterProgram;
var
  s1, s2: string;
begin
  itpt := TJvInterpreterProgram.Create(nil);
  try
    TJvAdapterAccess(itpt.SharedAdapter).Clear;
    itpt.SharedAdapter.AddGet(TMyForm, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TForm, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TMyChart, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TMyGrid, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TScrollingWinControl, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TWinControl, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TControl, 'Method', nil, 0, [], varEmpty);
    itpt.SharedAdapter.AddGet(TComponent, 'Method', nil, 0, [], varEmpty);

    MySort(itpt.SharedAdapter.GetList);
    s1 := DumpList(itpt.SharedAdapter.GetList);
    Check(Pos('TMyForm', s1) < Pos('TScrollingWinControl', s1));

  finally
    itpt.Free;
  end;
end;

initialization

RegisterTest(TTest1.ClassName, TTest1.Suite);

end.
